# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from simgrid import Actor, Engine, Host, Link, this_actor
import sys

# This example demonstrates how to attach a profile to a host or a link, to specify external changes to the resource speed.
# The first way to do so is to use a file in the XML, while the second is to use the programmatic interface.

def watcher():
  jupiter  = Host.by_name("Jupiter")
  fafard   = Host.by_name("Fafard")
  lilibeth = Host.by_name("Lilibeth")
  link1    = Link.by_name("1")
  link2    = Link.by_name("2")

  (links, lat) = jupiter.route_to(fafard)
  path = ""
  for l in links:
    path += ("" if len(path)==0 else ", ") + "link '" + l.name + "'"
  this_actor.info(f"Path from Jupiter to Fafard: {path} (latency: {lat:.6f}s).")

  for _ in range(10):
    this_actor.info("Fafard: %.0fMflops, Jupiter: %4.0fMflops, Lilibeth: %3.1fMflops, Link1: (%.2fMB/s %.0fms), Link2: (%.2fMB/s %.0fms)" % (
             fafard.speed * fafard.available_speed / 1000000,
             jupiter.speed * jupiter.available_speed / 1000000,
             lilibeth.speed * lilibeth.available_speed / 1000000, 
             link1.bandwidth / 1000, link1.latency * 1000, 
             link2.bandwidth / 1000, link2.latency * 1000))
    this_actor.sleep_for(1)

if __name__ == '__main__':
  e = Engine(sys.argv)
  # Load the platform description
  e.load_platform(sys.argv[1])

  # Add a new host programmatically, and attach a simple speed profile to it (alternate between full and half speed every two seconds
  lili = e.netzone_root.create_host("Lilibeth", 25e6)
  lili.set_speed_profile("""0 1.0
2 0.5""", 2)
  lili.seal()

  # Add a watcher of the changes
  Actor.create("watcher", Host.by_name("Fafard"), watcher)

  e.run()



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from simgrid import Actor,Comm, Engine, Host, Mailbox, this_actor
import sys

# This example shows how to block on the completion of a set of communications.
#
# As for the other asynchronous examples, the sender initiate all the messages it wants to send and
# pack the resulting simgrid.Comm objects in a list. All messages thus occur concurrently.
#
# The sender then blocks until all ongoing communication terminate, using simgrid.Comm.wait_all()


def sender(messages_count, msg_size, receivers_count):
     # List in which we store all ongoing communications
     pending_comms = []

     # Vector of the used mailboxes
     mboxes = [Mailbox.by_name("receiver-{:d}".format(i))
               for i in range(0, receivers_count)]

     # Start dispatching all messages to receivers, in a round robin fashion
     for i in range(0, messages_count):
         content = "Message {:d}".format(i)
         mbox = mboxes[i % receivers_count]

         this_actor.info("Send '{:s}' to '{:s}'".format(content, str(mbox)))

         # Create a communication representing the ongoing communication, and store it in pending_comms
         comm = mbox.put_async(content, msg_size)
         pending_comms.append(comm)

     # Start sending messages to let the workers know that they should stop
     for i in range(0, receivers_count):
         mbox = mboxes[i]
         this_actor.info("Send 'finalize' to '{:s}'".format(str(mbox)))
         comm = mbox.put_async("finalize", 0)
         pending_comms.append(comm)

     this_actor.info("Done dispatching all messages")

     # Now that all message exchanges were initiated, wait for their completion in one single call
     Comm.wait_all(pending_comms)

     this_actor.info("Goodbye now!")


def receiver(id):
    mbox = Mailbox.by_name("receiver-{:d}".format(id))

    this_actor.info("Wait for my first message")
    while True:
        received = mbox.get()
        this_actor.info("I got a '{:s}'.".format(received))
        if received == "finalize":
            break  # If it's a finalize message, we're done.


if __name__ == '__main__':
    e = Engine(sys.argv)

    # Load the platform description
    e.load_platform(sys.argv[1])

    Actor.create("sender", Host.by_name("Tremblay"), sender, 5, 1000000, 2)
    Actor.create("receiver", Host.by_name("Ruby"), receiver, 0)
    Actor.create("receiver", Host.by_name("Perl"), receiver, 1)

    e.run()



# Copyright (c) 2006-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

# This example shows how to declare and start your actors.
#
# The first step is to declare the code of your actors (what they do exactly does not matter to this example) and then
# you ask SimGrid to start your actors. There is three ways of doing so:
# - Directly, by instantiating your actor as parameter to Actor::create()
# - By first registering your actors before instantiating it;
# - Through the deployment file.
#
# This example shows all these solutions, even if you obviously should use only one of these solutions to start your
# actors. The most advised solution is to use a deployment file, as it creates a clear separation between your
# application and the settings to test it. This is a better scientific methodology. Actually, starting an actor with
# Actor.create() is mostly useful to start an actor from another actor.

from simgrid import Actor, Engine, Host, Mailbox, this_actor
import sys


def receiver(mailbox_name):
    """
    Our first class of actors is simply implemented with a function, that takes a single string as parameter.
    Later, this actor class is instantiated within the simulation.
    """
    mailbox = Mailbox.by_name(mailbox_name)

    this_actor.info(
        "Hello s4u, I'm ready to get any message you'd want on {:s}".format(mailbox.name))

    msg1 = mailbox.get()
    msg2 = mailbox.get()
    msg3 = mailbox.get()
    this_actor.info(
        "I received '{:s}', '{:s}' and '{:s}'".format(msg1, msg2, msg3))
    this_actor.info("I'm done. See you.")


def forwarder(*args):
    """Our second class of actors is also a function"""
    if len(args) < 2:
        raise AssertionError(
            "Actor forwarder requires 2 parameters, but got only {:d}".format(len(args)))
    mb_in = Mailbox.by_name(args[0])
    mb_out = Mailbox.by_name(args[1])

    msg = mb_in.get()
    this_actor.info("Forward '{:s}'.".format(msg))
    mb_out.put(msg, len(msg))


class Sender:
    """
    Declares a third class of actors which sends a message to the mailbox 'mb42'.
    The sent message is what was passed as parameter on creation (or 'GaBuZoMeu' by default)

    Later, this actor class is instantiated twice in the simulation.
    """

    def __init__(self, msg="GaBuZoMeu", mbox="mb42"):
        self.msg = msg
        self.mbox = mbox

    # Actors that are created as object will execute their __call__ method.
    # So, the following constitutes the main function of the Sender actor.
    def __call__(self):
        this_actor.info("Hello s4u, I have something to send")
        mailbox = Mailbox.by_name(self.mbox)

        mailbox.put(self.msg, len(self.msg))
        this_actor.info("I'm done. See you.")


if __name__ == '__main__':
    # Here comes the main function of your program

    # When your program starts, you have to first start a new simulation engine, as follows
    e = Engine(sys.argv)

    # Then you should load a platform file, describing your simulated platform
    e.load_platform("../../platforms/small_platform.xml")

    # And now you have to ask SimGrid to actually start your actors.
    #
    # The easiest way to do so is to implement the behavior of your actor in a single function,
    # as we do here for the receiver actors. This function can take any kind of parameters, as
    # long as the last parameters of Actor::create() match what your function expects.
    Actor.create("receiver", Host.by_name("Fafard"), receiver, "mb42")

    # If your actor is getting more complex, you probably want to implement it as a class instead,
    # as we do here for the sender actors. The main behavior goes into operator()() of the class.
    #
    # You can then directly start your actor, as follows:
    Actor.create("sender1", Host.by_name("Tremblay"), Sender())
    # If you want to pass parameters to your class, that's very easy: just use your constructors
    Actor.create("sender2", Host.by_name("Jupiter"), Sender("GloubiBoulga"))

    # But starting actors directly is considered as a bad experimental habit, since it ties the code
    # you want to test with the experimental scenario. Starting your actors from an external deployment
    # file in XML ensures that you can test your code in several scenarios without changing the code itself.
    #
    # For that, you first need to register your function or your actor as follows.
    e.register_actor("sender", Sender)
    e.register_actor("forwarder", forwarder)
    # Once actors and functions are registered, just load the deployment file
    e.load_deployment("actor-create_d.xml")

    # Once every actors are started in the engine, the simulation can start
    e.run()

    # Once the simulation is done, the program is ended



#! @PYTHON_EXECUTABLE@
# -*- coding: utf-8 -*-
"""

tesh -- testing shell
========================

Copyright (c) 2012-2022. The SimGrid Team. All rights reserved.

This program is free software; you can redistribute it and/or modify it
under the terms of the license (GNU LGPL) which comes with this package.

#TODO: child of child of child that printfs. Does it work?
#TODO: a child dies after its parent. What happen?

#TODO: regular expression in output
#ex: >> Time taken: [0-9]+s
#TODO: linked regular expression in output
#ex:
# >> Bytes sent: ([0-9]+)
# >> Bytes recv: \1
# then, even better:
# ! expect (\1 > 500)

"""

import sys
import os
import shlex
import re
import difflib
import signal
import argparse
import time

if sys.version_info[0] == 3:
    import subprocess
    import _thread
else:
    raise "This program is expected to run with Python3 only"

##############
#
# Utilities
#
#

def isWindows():
    return sys.platform.startswith('win')

# Singleton metaclass that works in Python 2 & 3
# http://stackoverflow.com/questions/6760685/creating-a-singleton-in-python

class _Singleton(type):
    """ A metaclass that creates a Singleton base class when called. """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class Singleton(_Singleton('SingletonMeta', (object,), {})):
    pass

SIGNALS_TO_NAMES_DICT = dict((getattr(signal, n), n)
                             for n in dir(signal) if n.startswith('SIG') and '_' not in n)

return_code = 0

# exit correctly
def tesh_exit(errcode):
    # If you do not flush some prints are skipped
    sys.stdout.flush()
    # os._exit exit even when executed within a thread
    os._exit(errcode)


def fatal_error(msg):
    print("[Tesh/CRITICAL] " + str(msg))
    tesh_exit(1)


# Set an environment variable.
# arg must be a string with the format "variable=value"
def setenv(arg):
    print("[Tesh/INFO] setenv " + arg)
    t = arg.split("=", 1)
    os.environ[t[0]] = t[1]
    # os.putenv(t[0], t[1]) does not work
    # see http://stackoverflow.com/questions/17705419/python-os-environ-os-putenv-usr-bin-env


# http://stackoverflow.com/questions/30734967/how-to-expand-environment-variables-in-python-as-bash-does
def expandvars2(path):
    return re.sub(r'(?<!\\)\$[A-Za-z_][A-Za-z0-9_]*', '', os.path.expandvars(path))

##############
#
# Cleanup on signal
#
#

# Global variable. Stores which process group should be killed (or None otherwise)
running_pids = list()

# Tests whether the process is dead already
def process_is_dead(pid):
    try:
        os.kill(pid, 0)
    except ProcessLookupError:
        return True
    except OSError as err:
        if err.errno == errno.ESRCH: # ESRCH == No such process. The process is now dead
            return True
    return False

# This function send TERM signal + KILL signal after 0.2s to the group of the specified process
def kill_process_group(pid):
    if pid is None:  # Nobody to kill. We don't know who to kill on windows, or we don't have anyone to kill on signal handler
        return

    try:
        pgid = os.getpgid(pid)
    except OSError:
        # os.getpgid failed. Ok, don't cleanup.
        return

    try:
        os.killpg(pgid, signal.SIGTERM)
        if process_is_dead(pid):
            return
        time.sleep(0.2)
        os.killpg(pgid, signal.SIGKILL)
    except OSError:
        # os.killpg failed. OK. Some subprocesses may still be running.
        pass

def signal_handler(signal, frame):
    print("Caught signal {}".format(SIGNALS_TO_NAMES_DICT[signal]))
    global running_pids
    running_pids_copy = running_pids # Just in case of interthread conflicts.
    for pid in running_pids_copy:
        kill_process_group(pid)
    running_pids.clear()
    tesh_exit(5)


##############
#
# Classes
#
#


# read file line per line (and concat line that ends with "\")
class FileReader(Singleton):
    def __init__(self, filename=None):
        if filename is None:
            self.filename = "(stdin)"
            self.f = sys.stdin
        else:
            self.filename_raw = filename
            self.filename = os.path.basename(filename)
            self.abspath = os.path.abspath(filename)
            self.f = open(self.filename_raw)

        self.linenumber = 0

    def __repr__(self):
        return self.filename + ":" + str(self.linenumber)

    def readfullline(self):
        try:
            line = next(self.f)
            self.linenumber += 1
        except StopIteration:
            return None
        if line[-1] == "\n":
            txt = line[0:-1]
        else:
            txt = line
        while len(line) > 1 and line[-2] == "\\":
            txt = txt[0:-1]
            line = next(self.f)
            self.linenumber += 1
            txt += line[0:-1]
        return txt


# keep the state of tesh (mostly configuration values)
class TeshState(Singleton):
    def __init__(self):
        self.threads = []
        self.args_suffix = ""
        self.ignore_regexps_common = []
        self.jenkins = False  # not a Jenkins run by default
        self.timeout = 10  # default value: 10 sec
        self.wrapper = None
        self.keep = False

    def add_thread(self, thread):
        self.threads.append(thread)

    def join_all_threads(self):
        for t in self.threads:
            t.acquire()
            t.release()

# Command line object


class Cmd(object):
    def __init__(self):
        self.input_pipe = []
        self.output_pipe_stdout = []
        self.output_pipe_stderr = []
        self.timeout = TeshState().timeout
        self.args = None
        self.linenumber = -1

        self.background = False
        # Python threads loose the cwd
        self.cwd = os.getcwd()

        self.ignore_output = False
        self.expect_return = [0]

        self.output_display = False

        self.sort = -1

        self.ignore_regexps = TeshState().ignore_regexps_common

    def add_input_pipe(self, l):
        self.input_pipe.append(l)

    def add_output_pipe_stdout(self, l):
        self.output_pipe_stdout.append(l)

    def add_output_pipe_stderr(self, l):
        self.output_pipe_stderr.append(l)

    def set_cmd(self, args, linenumber):
        self.args = args
        self.linenumber = linenumber

    def add_ignore(self, txt):
        self.ignore_regexps.append(re.compile(txt))

    def remove_ignored_lines(self, lines):
        for ign in self.ignore_regexps:
            lines = [l for l in lines if not ign.match(l)]
        return lines

    def _cmd_mkfile(self, argline):
        filename = argline[len("mkfile "):]
        file = open(filename, "w")
        if file is None:
            fatal_error("Unable to create file " + filename)
        file.write("\n".join(self.input_pipe))
        file.write("\n")
        file.close()

    def _cmd_cd(self, argline):
        args = shlex.split(argline)
        if len(args) != 2:
            fatal_error("Too many arguments to cd")
        try:
            os.chdir(args[1])
            print("[Tesh/INFO] change directory to " + args[1])
        except FileNotFoundError:
            print("Chdir to " + args[1] + " failed: No such file or directory")
            print("Test suite `" + FileReader().filename + "': NOK (system error)")
            tesh_exit(4)

    # Run the Cmd if possible.
    # Return False if nothing has been ran.

    def run_if_possible(self):
        if not self.can_run():
            return False
        if self.background:
            lock = _thread.allocate_lock()
            lock.acquire()
            TeshState().add_thread(lock)
            _thread.start_new_thread(Cmd._run, (self, lock))
        else:
            self._run()
        return True

    def _run(self, lock=None):
        # Python threads loose the cwd
        os.chdir(self.cwd)

        # retrocompatibility: support ${aaa:=.} variable format
        def replace_perl_variables(m):
            vname = m.group(1)
            vdefault = m.group(2)
            if vname in os.environ:
                return "$" + vname
            return vdefault

        self.args = re.sub(r"\${(\w+):=([^}]*)}", replace_perl_variables, self.args)

        # replace bash environment variables ($THINGS) to their values
        self.args = expandvars2(self.args)

        if re.match("^mkfile ", self.args) is not None:
            self._cmd_mkfile(self.args)
            if lock is not None:
                lock.release()
            return

        if re.match("^cd ", self.args) is not None:
            self._cmd_cd(self.args)
            if lock is not None:
                lock.release()
            return

        if TeshState().wrapper is not None:
            self.timeout *= 20
            self.args = TeshState().wrapper + self.args
        elif re.match(".*smpirun.*", self.args) is not None:
            self.args = "sh " + self.args
        if TeshState().jenkins and self.timeout is not None:
            self.timeout *= 10

        self.args += TeshState().args_suffix

        logs = list()
        logs.append("[{file}:{number}] {args}".format(file=FileReader().filename,
            number=self.linenumber, args=self.args))

        args = shlex.split(self.args)

        global running_pids
        local_pid = None
        global return_code

        try:
            preexec_function = None
            if not isWindows():
                preexec_function = lambda: os.setpgid(0, 0)
            proc = subprocess.Popen(
                args,
                bufsize=1,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                preexec_fn=preexec_function)
            if not isWindows():
                local_pid = proc.pid
                running_pids.append(local_pid)
        except PermissionError:
            logs.append("[{file}:{number}] Cannot start '{cmd}': The binary is not executable.".format(
                file=FileReader().filename, number=self.linenumber, cmd=args[0]))
            logs.append("[{file}:{number}] Current dir: {dir}".format(file=FileReader().filename,
                number=self.linenumber, dir=os.getcwd()))
            return_code = max(3, return_code)
            print('\n'.join(logs))
            return
        except NotADirectoryError:
            logs.append("[{file}:{number}] Cannot start '{cmd}': The path to binary does not exist.".format(
                file=FileReader().filename, number=self.linenumber, cmd=args[0]))
            logs.append("[{file}:{number}] Current dir: {dir}".format(file=FileReader().filename,
                number=self.linenumber, dir=os.getcwd()))
            return_code = max(3, return_code)
            print('\n'.join(logs))
            return
        except FileNotFoundError:
            logs.append("[{file}:{number}] Cannot start '{cmd}': File not found.".format(
                file=FileReader().filename, number=self.linenumber, cmd=args[0]))
            return_code = max(3, return_code)
            print('\n'.join(logs))
            return
        except OSError as osE:
            if osE.errno == 8:
                osE.strerror += "\nOSError: [Errno 8] Executed scripts should start with shebang line (like #!/usr/bin/env sh)"
            raise osE

        cmdName = FileReader().filename + ":" + str(self.linenumber)
        try:
            (stdout_data, stderr_data) = proc.communicate("\n".join(self.input_pipe), self.timeout)
            timeout_reached = False
        except subprocess.TimeoutExpired:
            timeout_reached = True
            logs.append("Test suite `{file}': NOK (<{cmd}> timeout after {timeout} sec)".format(
                file=FileReader().filename, cmd=cmdName, timeout=self.timeout))
            running_pids.remove(local_pid)
            kill_process_group(local_pid)
            # Try to get the output of the timeout process, to help in debugging.
            try:
                (stdout_data, stderr_data) = proc.communicate(timeout=1)
            except subprocess.TimeoutExpired:
                logs.append("[{file}:{number}] Could not retrieve output. Killing the process group failed?".format(
                    file=FileReader().filename, number=self.linenumber))
                return_code = max(3, return_code)
                print('\n'.join(logs))
                return

        if self.output_display:
            logs.append(str(stdout_data))

        # remove text colors
        ansi_escape = re.compile(r'\x1b[^m]*m')
        stdout_data = ansi_escape.sub('', stdout_data)

        if self.ignore_output:
            logs.append("(ignoring the output of <{cmd}> as requested)".format(cmd=cmdName))
        else:
            stdouta = stdout_data.split("\n")
            while stdouta and stdouta[-1] == "":
                del stdouta[-1]
            stdouta = self.remove_ignored_lines(stdouta)
            stdcpy = stdouta[:]

            # Mimic the "sort" bash command, which is case unsensitive.
            if self.sort == 0:
                stdouta.sort(key=lambda x: x.lower())
                self.output_pipe_stdout.sort(key=lambda x: x.lower())
            elif self.sort > 0:
                stdouta.sort(key=lambda x: x[:self.sort].lower())
                self.output_pipe_stdout.sort(key=lambda x: x[:self.sort].lower())

            diff = list(
                difflib.unified_diff(
                    self.output_pipe_stdout,
                    stdouta,
                    lineterm="",
                    fromfile='expected',
                    tofile='obtained'))
            if diff:
                logs.append("Output of <{cmd}> mismatch:".format(cmd=cmdName))
                if self.sort >= 0:  # If sorted, truncate the diff output and show the unsorted version
                    difflen = 0
                    for line in diff:
                        if difflen < 50:
                            print(line)
                        difflen += 1
                    if difflen > 50:
                        logs.append("(diff truncated after 50 lines)")
                    logs.append("Unsorted observed output:\n")
                    for line in stdcpy:
                        logs.append(line)
                else:  # If not sorted, just display the diff
                    for line in diff:
                        logs.append(line)

                logs.append("Test suite `{file}': NOK (<{cmd}> output mismatch)".format(
                    file=FileReader().filename, cmd=cmdName))
                if lock is not None:
                    lock.release()
                if TeshState().keep:
                    f = open('obtained', 'w')
                    obtained = stdout_data.split("\n")
                    while obtained and obtained[-1] == "":
                        del obtained[-1]
                    obtained = self.remove_ignored_lines(obtained)
                    for line in obtained:
                        f.write("> " + line + "\n")
                    f.close()
                    logs.append("Obtained output kept as requested: {path}".format(path=os.path.abspath("obtained")))
                return_code = max(2, return_code)
                print('\n'.join(logs))
                return

        if timeout_reached:
            return_code = max(3, return_code)
            print('\n'.join(logs))
            return

        if not proc.returncode in self.expect_return:
            if proc.returncode >= 0:
                logs.append("Test suite `{file}': NOK (<{cmd}> returned code {code})".format(
                    file=FileReader().filename, cmd=cmdName, code=proc.returncode))
                if lock is not None:
                    lock.release()
                return_code = max(2, return_code)
                print('\n'.join(logs))
                return

            logs.append("Test suite `{file}': NOK (<{cmd}> got signal {sig})".format(
                file=FileReader().filename, cmd=cmdName,
                sig=SIGNALS_TO_NAMES_DICT[-proc.returncode]))
            if lock is not None:
                lock.release()
            return_code = max(max(-proc.returncode, 1), return_code)
            print('\n'.join(logs))
            return

        if lock is not None:
            lock.release()

        print('\n'.join(logs))

    def can_run(self):
        return self.args is not None

##############
#
# Main
#
#

if __name__ == '__main__':
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    parser = argparse.ArgumentParser(description='tesh -- testing shell')
    group1 = parser.add_argument_group('Options')
    group1.add_argument('teshfile', nargs='?', help='Name of teshfile, stdin if omitted')
    group1.add_argument(
        '--cd',
        metavar='some/directory',
        help='ask tesh to switch the working directory before launching the tests')
    group1.add_argument('--setenv', metavar='var=value', action='append', help='set a specific environment variable')
    group1.add_argument('--cfg', metavar='arg', action='append', help='add parameter --cfg=arg to each command line')
    group1.add_argument('--log', metavar='arg', action='append', help='add parameter --log=arg to each command line')
    group1.add_argument(
        '--ignore-jenkins',
        action='store_true',
        help='ignore all cruft generated on SimGrid continuous integration servers')
    group1.add_argument('--wrapper', metavar='arg', help='Run each command in the provided wrapper (eg valgrind)')
    group1.add_argument(
        '--keep',
        action='store_true',
        help='Keep the obtained output when it does not match the expected one')

    options = parser.parse_args()

    if options.cd is not None:
        print("[Tesh/INFO] change directory to " + options.cd)
        os.chdir(options.cd)

    if options.ignore_jenkins:
        print("Ignore all cruft seen on SimGrid's continuous integration servers")
        # Note: regexps should match at the beginning of lines
        TeshState().ignore_regexps_common = [
            re.compile(r"profiling:"),
            re.compile(r"Unable to clean temporary file C:"),
            re.compile(r".*Configuration change: Set 'contexts/"),
            re.compile(r"Picked up JAVA_TOOL_OPTIONS: "),
            re.compile(r"Picked up _JAVA_OPTIONS: "),
            re.compile(r"==[0-9]+== ?WARNING: ASan doesn't fully support"),
            re.compile(r"==[0-9]+== ?WARNING: ASan is ignoring requested __asan_handle_no_return: stack "),
            re.compile(r"False positive error reports may follow"),
            re.compile(r"For details see http://code\.google\.com/p/address-sanitizer/issues/detail\?id=189"),
            re.compile(r"For details see https://github\.com/google/sanitizers/issues/189"),
            re.compile(r"Python runtime initialized with LC_CTYPE=C .*"),
            # Seen on CircleCI
            re.compile(r"cmake: /usr/local/lib/libcurl\.so\.4: no version information available \(required by cmake\)"),
            re.compile(r".*mmap broken on FreeBSD, but dlopen\+thread broken too\. Switching to dlopen\+raw contexts\."),
            re.compile(r".*dlopen\+thread broken on Apple and BSD\. Switching to raw contexts\."),
        ]
        TeshState().jenkins = True  # This is a Jenkins build

    if options.teshfile is None:
        f = FileReader(None)
        print("Test suite from stdin")
    else:
        if not os.path.isfile(options.teshfile):
            print("Cannot open teshfile '" + options.teshfile + "': File not found")
            tesh_exit(3)
        f = FileReader(options.teshfile)
        print("Test suite '" + f.abspath + "'")

    if options.setenv is not None:
        for e in options.setenv:
            setenv(e)

    if options.cfg is not None:
        for c in options.cfg:
            TeshState().args_suffix += " --cfg=" + c
    if options.log is not None:
        for l in options.log:
            TeshState().args_suffix += " --log=" + l

    if options.wrapper is not None:
        TeshState().wrapper = options.wrapper

    if options.keep:
        TeshState().keep = True

    # cmd holds the current command line
    # tech commands will add some parameters to it
    # when ready, we execute it.
    cmd = Cmd()

    line = f.readfullline()
    while line is not None:
        # print(">>============="+line+"==<<")
        if not line:
            #print ("END CMD block")
            if cmd.run_if_possible():
                cmd = Cmd()

        elif line[0] == "#":
            pass

        elif line[0:2] == "p ":
            print("[" + str(FileReader()) + "] " + line[2:])

        elif line[0:2] == "< ":
            cmd.add_input_pipe(line[2:])
        elif line[0:1] == "<":
            cmd.add_input_pipe(line[1:])

        elif line[0:2] == "> ":
            cmd.add_output_pipe_stdout(line[2:])
        elif line[0:1] == ">":
            cmd.add_output_pipe_stdout(line[1:])

        elif line[0:2] == "$ ":
            if cmd.run_if_possible():
                cmd = Cmd()
            cmd.set_cmd(line[2:], f.linenumber)

        elif line[0:2] == "& ":
            if cmd.run_if_possible():
                cmd = Cmd()
            cmd.set_cmd(line[2:], f.linenumber)
            cmd.background = True

        elif line[0:15] == "! output ignore":
            cmd.ignore_output = True
            #print("cmd.ignore_output = True")
        elif line[0:16] == "! output display":
            cmd.output_display = True
            cmd.ignore_output = True
        elif line[0:15] == "! expect return":
            cmd.expect_return = [int(line[16:])]
            #print("expect return "+str(int(line[16:])))
        elif line[0:15] == "! expect signal":
            cmd.expect_return = []
            for sig in (line[16:]).split("|"):
                # get the signal integer value from the signal module
                if sig not in signal.__dict__:
                    fatal_error("unrecognized signal '" + sig + "'")
                sig = int(signal.__dict__[sig])
                # popen return -signal when a process ends with a signal
                cmd.expect_return.append(-sig)
        elif line[0:len("! timeout ")] == "! timeout ":
            if "no" in line[len("! timeout "):]:
                cmd.timeout = None
            else:
                cmd.timeout = int(line[len("! timeout "):])

        elif line[0:len("! output sort")] == "! output sort":
            if len(line) >= len("! output sort "):
                sort = int(line[len("! output sort "):])
            else:
                sort = 0
            cmd.sort = sort
        elif line[0:len("! setenv ")] == "! setenv ":
            setenv(line[len("! setenv "):])

        elif line[0:len("! ignore ")] == "! ignore ":
            cmd.add_ignore(line[len("! ignore "):])

        else:
            fatal_error("UNRECOGNIZED OPTION")

        line = f.readfullline()

    cmd.run_if_possible()

    TeshState().join_all_threads()

    if return_code == 0:
        if f.filename == "(stdin)":
            print("Test suite from stdin OK")
        else:
            print("Test suite `" + f.filename + "' OK")
    tesh_exit(return_code)



# Copyright (c) 2006-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

# This example shows how to simulate a non-linear resource sharing for
# network links.

from simgrid import Actor, Engine, Comm, Mailbox, NetZone, Link, LinkInRoute, this_actor
import sys
import functools

class Sender:
  """
  Send a series of messages to mailbox "receiver"
  """
  def __init__(self, msg_count: int, msg_size=int(1e6)):
    self.msg_count = msg_count
    self.msg_size = msg_size

  # Actors that are created as object will execute their __call__ method.
  # So, the following constitutes the main function of the Sender actor.
  def __call__(self):
    pending_comms = []
    mbox = Mailbox.by_name("receiver")

    for i in range(self.msg_count):
      msg = "Message " + str(i)
      size = self.msg_size * (i + 1)
      this_actor.info("Send '%s' to '%s, msg size: %d'" % (msg, mbox.name, size))
      comm = mbox.put_async(msg, size)
      pending_comms.append(comm)

    this_actor.info("Done dispatching all messages")

    # Now that all message exchanges were initiated, wait for their completion in one single call
    Comm.wait_all(pending_comms)

    this_actor.info("Goodbye now!")

class Receiver:
  """
  Receiver actor: wait for N messages on the mailbox "receiver"
  """

  def __init__(self, msg_count=10):
    self.msg_count = msg_count

  def __call__(self):
    mbox = Mailbox.by_name("receiver")

    pending_msgs = []
    pending_comms = []

    this_actor.info("Wait for %d messages asynchronously" % self.msg_count)
    for _ in range(self.msg_count):
      comm, data = mbox.get_async()
      pending_comms.append(comm)
      pending_msgs.append(data)

    while len(pending_comms) > 0:
      index = Comm.wait_any(pending_comms)
      msg = pending_msgs[index].get()
      this_actor.info("I got '%s'." % msg)
      del pending_comms[index]
      del pending_msgs[index]

####################################################################################################
def link_nonlinear(link: Link, capacity: float, n: int) -> float:
  """
  Non-linear resource sharing for links

  Note that the callback is called twice in this example:
   1) link UP: with the number of active flows (from 9 to 1)
   2) link DOWN: with 0 active flows. A crosstraffic communication is happing
   in the down link, but it's not considered as an active flow.
  """
  # emulates a degradation in link according to the number of flows
  # you probably want something more complex than that and based on real
  # experiments
  capacity = min(capacity, capacity * (1.0 - (n - 1) / 10.0))
  this_actor.info("Link %s, %d active communications, new capacity %f" % (link.name, n, capacity))
  return capacity

def load_platform():
  """
  Create a simple 2-hosts platform */
   ________                 __________
  | Sender |===============| Receiver |
  |________|    Link1      |__________|

  """
  zone = NetZone.create_full_zone("Zone1")
  sender = zone.create_host("sender", 1).seal()
  receiver = zone.create_host("receiver", 1).seal()

  link = zone.create_split_duplex_link("link1", 1e6)
  # setting same callbacks (could be different) for link UP/DOWN in split-duplex link
  link.link_up.set_sharing_policy(
      Link.SharingPolicy.NONLINEAR,
      functools.partial(link_nonlinear, link.link_up))
  link.link_down.set_sharing_policy(
      Link.SharingPolicy.NONLINEAR,
      functools.partial(link_nonlinear, link.link_down))
  link.set_latency(10e-6).seal()

  # create routes between nodes
  zone.add_route(sender.netpoint, receiver.netpoint, None, None,
                 [LinkInRoute(link, LinkInRoute.Direction.UP)], True)
  zone.seal()

  # create actors Sender/Receiver
  Actor.create("receiver", receiver, Receiver(9))
  Actor.create("sender", sender, Sender(9))

###################################################################################################

if __name__ == '__main__':
  e = Engine(sys.argv)

  # create platform
  load_platform()

  # runs the simulation
  e.run()

  # explicitly deleting Engine object to avoid segfault during cleanup phase.
  # During Engine destruction, the cleanup of std::function linked to link_non_linear callback is called.
  # If we let the cleanup by itself, it fails trying on its destruction because the python main program
  # has already freed its variables
  del(e)



# Copyright (c) 2019-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

# First failing example for bug #9 on Framagit (Python bindings crashing)
#
# A heavy operation is added after e.run() (here, it's "import gc"). Note that
# the failure only occurs when the actor "sender" starts first, and the
# scheduling is interlaced: sender starts, receiver starts, sender terminates,
# receiver terminates.

import sys
from simgrid import Engine, this_actor

def sender():
    this_actor.sleep_for(3)
    this_actor.info("Goodbye now!")

def receiver():
    this_actor.sleep_for(5)
    this_actor.info("Five seconds elapsed")

if __name__ == '__main__':
    e = Engine(sys.argv)

    e.load_platform(sys.argv[1])             # Load the platform description

    # Register the classes representing the actors
    e.register_actor("sender", sender)
    e.register_actor("receiver", receiver)

    e.load_deployment(sys.argv[2])

    e.run()
    this_actor.info("Dummy import...")
    import gc
    gc.collect()
    this_actor.info("done.")



#!/usr/bin/env python3

# Copyright (c) 2018-2022. The SimGrid Team. All rights reserved.

# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

'''
This script is intended to convert SMPI time independent traces (TIT) from the
old format (simgrid version <= 3.19) to the new format.

On the previous format each MPI_wait calls were associated to the last ISend of
IRecv call arbitrarily.

This new that includes tags field that links MPI_wait calls to the
MPI_ISend or MPI_IRecv associated to this wait.

This script reproduce the old behavior of simgrid because information are
missing to add the tags properly. It also lower case all the mpi calls.

It takes in input (as argument or in stdin) the trace list file that is only a
simple text file that contains path of actual TIT files split by rank.

It creates a directory to put the traces on ("converted_traces" by default)
'''

import os
import pathlib
import shutil


def insert_elem(split_line, position, elem):
    split_line.insert(position, elem)
    return " ".join(split_line)


def convert_trace(trace_path, base_path, output_path, trace_version="1.0"):
    old_file_path = os.path.join(base_path, trace_path)
    with open(old_file_path) as trace_file:

        new_file_path = os.path.join(output_path, trace_path)
        pathlib.Path(os.path.dirname(new_file_path)).mkdir(
            parents=True, exist_ok=True)

        with open(new_file_path, "w") as new_trace:
            # Write header
            new_trace.write("# version: " + trace_version + "\n")

            last_async_call_src = None
            last_async_call_dst = None
            for line_num, line in enumerate(trace_file.readlines()):
                # print(line)
                new_line = None
                split_line = line.lower().strip().split(" ")
                mpi_call = split_line[1]

                if mpi_call == "recv" or mpi_call == "send":
                    new_line = insert_elem(split_line, 3, "0")

                elif mpi_call == "irecv" or mpi_call == "isend":
                    last_async_call_src = split_line[0]
                    last_async_call_dst = split_line[2]
                    new_line = insert_elem(split_line, 3, "0")
                    # print("found async call in line: "+ str(line_num))

                elif mpi_call == "wait":
                    # print("found wait call in line: "+ str(line_num))
                    if (last_async_call_src is None
                            or last_async_call_dst is None):
                        raise Exception("Invalid traces: No Isend or Irecv "
                                        "found before the wait in line " +
                                        str(line_num) + " in file " + old_file_path)
                    insert_elem(split_line, 2, last_async_call_src)
                    insert_elem(split_line, 3, last_async_call_dst)
                    new_line = insert_elem(split_line, 4, "0")

                if new_line is not None:
                    # print("line: "+ str(line_num) + " in file " + old_file_path +
                    #        " processed\n:OLD: " + line + "\n:NEW: " + new_line)
                    new_trace.write(new_line + "\n")
                else:
                    new_trace.write(line.lower())


if __name__ == "__main__":
    import argparse
    import sys

    parser = argparse.ArgumentParser(description=sys.modules[__name__].__doc__)

    parser.add_argument('trace_list_file', type=argparse.FileType('r'),
                        default=sys.stdin, help="The trace list file (e.g. smpi_simgrid.txt)")

    parser.add_argument('--output_path', '-o', default="converted_traces",
                        help="The path where converted traces will be put")

    args = parser.parse_args()

    trace_list_file_path = args.trace_list_file.name

    # creates results dir
    pathlib.Path(args.output_path).mkdir(
        parents=True, exist_ok=True)

    # copy trace list file
    try:
        shutil.copy(trace_list_file_path, args.output_path)
    except shutil.SameFileError:
        print("ERROR: Inplace replacement of the trace is not supported: "
              "Please, select another output path")
        sys.exit(-1)

    with open(trace_list_file_path) as tracelist_file:
        trace_list = tracelist_file.readlines()

    # get based path relative to trace list file
    base_path = os.path.dirname(trace_list_file_path)

    trace_list = [x.strip() for x in trace_list]

    # process trace files
    for trace_path in trace_list:
        if os.path.isabs(trace_path):
            sys.exit("ERROR: Absolute path in the trace list file is not supported")
        convert_trace(trace_path, base_path, args.output_path)

    print("Traces converted!")
    print("Result directory:\n" + args.output_path)



# -*- coding: utf-8 -*-

# Useful doc: https://www.sphinx-doc.org/en/master/extdev/markupapi.html
# Example: https://www.sphinx-doc.org/en/master/development/tutorials/recipe.html

import os
from docutils.parsers.rst import Directive, directives
from docutils import nodes
from docutils.statemachine import StringList
from sphinx.util.osutil import copyfile
from sphinx.util import logging

CSS_FILE = 'showfile.css'
JS_FILE = 'showfile.js'

class ShowFileDirective(Directive):
    """
    Show a file or propose it to download.
    """

    has_content = False
    optional_arguments = 1
    option_spec = {
        'language': directives.unchanged
    }

    def run(self):

        filename = self.arguments[0]
        language = "python"
        if 'language' in self.options:
            language = self.options['language']

        logger = logging.getLogger(__name__)
#        logger.info('showfile {} in {}'.format(filename, language))

        new_content = [
          '.. toggle-header::',
          '   :header: View {}'.format(filename),
          '',
          '   `Download {} <https://framagit.org/simgrid/simgrid/tree/master/{}>`_'.format(os.path.basename(filename), filename),
          '',
          '   .. literalinclude:: ../../{}'.format(filename),
          '      :language: {}'.format(language),
          ''
        ]

        for idx, line in enumerate(new_content):
#            logger.info('{} {}'.format(idx,line))
            self.content.data.insert(idx, line)
            self.content.items.insert(idx, (None, idx))

        node = nodes.container()
        self.state.nested_parse(self.content, self.content_offset, node)
        return node.children

class ExampleTabDirective(Directive):
    """
    A group-tab for a given language, in the presentation of the examples.
    """
    has_content = True
    optional_arguments = 0
    mandatory_argument = 0

    def run(self):
        self.assert_has_content()

        filename = self.content[0].strip()
        self.content.trim_start(1)

        (language, langcode) = (None, None)
        if filename[-3:] == '.py':
            language = 'Python'
            langcode = 'py'
        elif filename[-2:] == '.c':
            language = 'C'
            langcode = 'c'
        elif filename[-4:] == '.cpp':
            language = 'C++'
            langcode = 'cpp'
        elif filename[-4:] == '.xml':
            language = 'XML'
            langcode = 'xml'
        else:
            raise Exception("Unknown language '{}'. Please choose '.cpp', '.py', '.c' or '.xml'".format(language))

        for idx, line in enumerate(self.content.data):
            self.content.data[idx] = '   ' + line

        for idx, line in enumerate([
            '.. group-tab:: {}'.format(language),
            '   ']):
            self.content.data.insert(idx, line)
            self.content.items.insert(idx, (None, idx))

        for line in [
            '',
            '   .. showfile:: {}'.format(filename),
            '      :language: {}'.format(langcode),
            '']:
            idx = len(self.content.data)
            self.content.data.insert(idx, line)
            self.content.items.insert(idx, (None, idx))

#        logger = logging.getLogger(__name__)
#        logger.info('------------------')
#        for line in self.content.data:
#            logger.info('{}'.format(line))

        node = nodes.container()
        self.state.nested_parse(self.content, self.content_offset, node)
        return node.children

class ToggleDirective(Directive):
    has_content = True
    option_spec = {
        'header': directives.unchanged,
        'show': directives.flag
    }
    optional_arguments = 1

    def run(self):
        node = nodes.container()
        node['classes'].append('toggle-content')
        if "show" not in self.options:
            # This :show: thing is not working, and I fail to see why.
            # Only the hidden-content class gets a call to hide() in the Javascript,
            # and :show:n block# still get hidden when I load the page.
            # No idea what's going on (Mt)
            node['classes'].append('hidden-content')

        par = nodes.container()
        par['classes'].append('toggle-header')
        if self.arguments and self.arguments[0]:
            par['classes'].append(self.arguments[0])

        self.state.nested_parse(StringList([self.options["header"]]), self.content_offset, par)
        self.state.nested_parse(self.content, self.content_offset, node)

        return [par, node]

def add_assets(app):
    app.add_css_file(CSS_FILE)
    app.add_js_file(JS_FILE)


def copy_assets(app, exception):
    if app.builder.name not in ['html', 'readthedocs'] or exception:
        return
    logger = logging.getLogger(__name__)
    logger.info('Copying showfile stylesheet/javascript... ', nonl=True)
    dest = os.path.join(app.builder.outdir, '_static', CSS_FILE)
    source = os.path.join(os.path.abspath(os.path.dirname(__file__)), CSS_FILE)
    copyfile(source, dest)
    dest = os.path.join(app.builder.outdir, '_static', JS_FILE)
    source = os.path.join(os.path.abspath(os.path.dirname(__file__)), JS_FILE)
    copyfile(source, dest)
    logger.info('done')

def setup(app):
    app.add_directive('toggle-header', ToggleDirective)
    app.add_directive('showfile', ShowFileDirective)
    app.add_directive('example-tab', ExampleTabDirective)

    app.connect('builder-inited', add_assets)
    app.connect('build-finished', copy_assets)



# Copyright (c) 2006-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

# This example shows how to simulate a non-linear resource sharing for
# CPUs.


from simgrid import Actor, Engine, NetZone, Host, this_actor
import sys
import functools


def runner():
    computation_amount = this_actor.get_host().speed
    n_task = 10

    this_actor.info("Execute %d tasks of %g flops, should take %d second in a CPU without degradation. It will take the double here." % (
        n_task, computation_amount, n_task))
    tasks = [this_actor.exec_init(computation_amount).start()
             for _ in range(n_task)]

    this_actor.info("Waiting for all tasks to be done!")
    for task in tasks:
        task.wait()

    this_actor.info("Finished executing. Goodbye now!")


def cpu_nonlinear(host: Host, capacity: float, n: int) -> float:
    """ Non-linear resource sharing for CPU """
    # emulates a degradation in CPU according to the number of tasks
    # totally unrealistic but for learning purposes
    capacity = capacity / 2 if n > 1 else capacity
    this_actor.info("Host %s, %d concurrent tasks, new capacity %f" %
                    (host.name, n, capacity))
    return capacity


def load_platform():
    """ Create a simple 1-host platform """
    zone = NetZone.create_empty_zone("Zone1")
    runner_host = zone.create_host("runner", 1e6)
    runner_host.set_sharing_policy(
        Host.SharingPolicy.NONLINEAR, functools.partial(cpu_nonlinear, runner_host))
    runner_host.seal()
    zone.seal()

    # create actor runner
    Actor.create("runner", runner_host, runner)


if __name__ == '__main__':
    e = Engine(sys.argv)

    # create platform
    load_platform()

    # runs the simulation
    e.run()

    # explicitly deleting Engine object to avoid segfault during cleanup phase.
    # During Engine destruction, the cleanup of std::function linked to non_linear callback is called.
    # If we let the cleanup by itself, it fails trying on its destruction because the python main program
    # has already freed its variables
    del(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2014-2022. The SimGrid Team. All rights reserved.

# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

import re
import glob
import sys

class Arg(object):

    def __init__(self, name, thetype):
        self.name = name
        self.type = thetype

    def field(self):
        return self.simcall_types[self.type]

    def rettype(self):
        return self.type


class Simcall(object):
    simcalls_body = None
    simcalls_pre = None

    def __init__(self, name, handler, res, args, call_kind):
        self.name = name
        self.res = res
        self.args = args
        self.need_handler = handler
        self.call_kind = call_kind

    def check(self):
        # libsmx.c  simcall_BODY_
        if self.simcalls_body is None:
            f = open('libsmx.cpp')
            self.simcalls_body = set(re.findall(r'simcall_BODY_(.*?)\(', f.read()))
            f.close()
        if self.name not in self.simcalls_body:
            print ('# ERROR: No function calling simcall_BODY_%s' % self.name)
            print ('# Add something like this to libsmx.c:')
            print ('%s simcall_%s(%s)' % (self.res.rettype(), self.name, ', '.
                                          join('%s %s' % (arg.rettype(), arg.name) for arg in self.args)))
            print ('{')
            print ('  return simcall_BODY_%s(%s);' % (self.name, "..."))
            print ('}')
            return False

        # smx_*.c void simcall_HANDLER_host_on(smx_simcall_t simcall,
        # smx_host_t h)
        if self.simcalls_pre is None:
            self.simcalls_pre = set()
            for fn in glob.glob('smx_*') + glob.glob('../kernel/actor/ActorImpl*') + \
                    glob.glob('../mc/*cpp') + glob.glob('../kernel/activity/*cpp'):
                f = open(fn)
                self.simcalls_pre |= set(re.findall(r'simcall_HANDLER_(.*?)\(', f.read()))
                f.close()
        if self.need_handler:
            if self.name not in self.simcalls_pre:
                print ('# ERROR: No function called simcall_HANDLER_%s' % self.name)
                print ('# Add something like this to the relevant C file (like smx_io.c if it\'s an IO call):')
                print ('%s simcall_HANDLER_%s(smx_simcall_t simcall%s)' % (self.res.rettype(), self.name, ''.
                                                                           join(', %s %s' % (arg.rettype(), arg.name)for arg in self.args)))
                print ('{')
                print ('  // Your code handling the simcall')
                print ('}')
                return False
        else:
            if self.name in self.simcalls_pre:
                print (
                    '# ERROR: You have a function called simcall_HANDLER_%s, but that simcall is not using any handler' %
                    self.name)
                print ('# Either change your simcall definition, or kill that function')
                return False
        return True

    def enum(self):
        return '  %s,' % (self.name.upper())

    def string(self):
        return '    "Simcall::%s",' % self.name.upper()

    def accessors(self):
        res = []
        res.append('')
        regex = re.compile(r"^boost::intrusive_ptr<(.*?)>(.*)$")  # to compute the raw type
        # Arguments getter/setters
        for i in range(len(self.args)):
            arg = self.args[i]
            rawtype = regex.sub(r'\1*\2', arg.rettype())
            res.append('static inline %s simcall_%s__get__%s(smx_simcall_t simcall)' % (
                arg.rettype(), self.name, arg.name))
            res.append('{')
            res.append('  return simgrid::simix::unmarshal<%s>(simcall->args_[%i]);' % (arg.rettype(), i))
            res.append('}')
            res.append('static inline %s simcall_%s__getraw__%s(smx_simcall_t simcall)' % (
                rawtype, self.name, arg.name))
            res.append('{')
            res.append('  return simgrid::simix::unmarshal_raw<%s>(simcall->args_[%i]);' % (rawtype, i))
            res.append('}')
            res.append('static inline void simcall_%s__set__%s(smx_simcall_t simcall, %s arg)' % (
                self.name, arg.name, arg.rettype()))
            res.append('{')
            res.append('  simgrid::simix::marshal<%s>(simcall->args_[%i], arg);' % (arg.rettype(), i))
            res.append('}')

        # Return value getter/setters
        if self.res.type != 'void':
            rawtype = regex.sub(r'\1*\2', self.res.rettype())
            res.append(
                'static inline %s simcall_%s__get__result(smx_simcall_t simcall)' % (self.res.rettype(), self.name))
            res.append('{')
            res.append('  return simgrid::simix::unmarshal<%s>(simcall->result_);' % self.res.rettype())
            res.append('}')
            res.append('static inline %s simcall_%s__getraw__result(smx_simcall_t simcall)' % (rawtype, self.name))
            res.append('{')
            res.append('  return simgrid::simix::unmarshal_raw<%s>(simcall->result_);' % rawtype)
            res.append('}')
            res.append(
                'static inline void simcall_%s__set__result(smx_simcall_t simcall, %s result)' % (self.name, self.res.rettype()))
            res.append('{')
            res.append('  simgrid::simix::marshal<%s>(simcall->result_, result);' % (self.res.rettype()))
            res.append('}')
        return '\n'.join(res)

    def case(self):
        res = []
        indent = '    '
        args = ["simgrid::simix::unmarshal<%s>(simcall_.args_[%d])" % (arg.rettype(), i)
                for i, arg in enumerate(self.args)]
        res.append(indent + 'case Simcall::%s:' % (self.name.upper()))
        if self.need_handler:
            call = "simcall_HANDLER_%s(&simcall_%s%s)" % (self.name,
                                                        ", " if args else "",
                                                        ', '.join(args))
        else:
            call = "SIMIX_%s(%s)" % (self.name, ', '.join(args))
        if self.call_kind == 'Func':
            res.append(indent + "  simgrid::simix::marshal<%s>(simcall_.result_, %s);" % (self.res.rettype(), call))
        else:
            res.append(indent + "  " + call + ";")
        if self.call_kind != 'Blck':
            res.append(indent + '  simcall_answer();')
        res.append(indent + '  break;')
        res.append('')
        return '\n'.join(res)

    def body(self):
        res = ['']
        res.append(
            'inline static %s simcall_BODY_%s(%s)' % (self.res.rettype(),
                                                      self.name,
                                                      ', '.join('%s %s' % (arg.rettype(), arg.name) for arg in self.args)))
        res.append('{')
        res.append('  if (false) /* Go to that function to follow the code flow through the simcall barrier */')
        if self.need_handler:
            res.append('    simcall_HANDLER_%s(%s);' % (self.name,
                                                        ', '.join(["&simgrid::kernel::actor::ActorImpl::self()->simcall_"] + [arg.name for arg in self.args])))
        else:
            res.append('    SIMIX_%s(%s);' % (self.name,
                                              ', '.join(arg.name for arg in self.args)))
        res.append('  return simcall<%s%s>(Simcall::%s%s);' % (
            self.res.rettype(),
            "".join([", " + arg.rettype() for i, arg in enumerate(self.args)]),
            self.name.upper(),
            "".join([", " + arg.name for i, arg in enumerate(self.args)])
        ))
        res.append('}')
        return '\n'.join(res)

    def handler_prototype(self):
        if self.need_handler:
            return "XBT_PRIVATE %s simcall_HANDLER_%s(smx_simcall_t simcall%s);" % (self.res.rettype() if self.call_kind == 'Func' else 'void',
                                                                                    self.name,
                                                                                    ''.join(', %s %s' % (arg.rettype(), arg.name)
                                                                                            for i, arg in enumerate(self.args)))
        return ""


def parse(fn):
    simcalls = []
    resdi = None
    simcalls_guarded = {}
    for line in open(fn).read().split('\n'):
        if line.startswith('##'):
            resdi = []
            simcalls_guarded[re.search(r'## *(.*)', line).group(1)] = resdi
        if line.startswith('#') or not line:
            continue
        match = re.match(
            r'^(\S+)\s+([^\)\(\s]+)\s*\(*(.*)\)\s*(\[\[.*\]\])?\s*;\s*$', line)
        if not match:
            raise AssertionError(line)
        ret, name, args, attrs = match.groups()
        sargs = []
        if not re.match(r"^\s*$", args):
            for arg in re.split(",", args):
                args = args.strip()
                match = re.match(r"^(.*?)\s*?(\S+)$", arg)
                t, n = match.groups()
                t = t.strip()
                n = n.strip()
                sargs.append(Arg(n, t))
        if ret == "void":
            ans = "Proc"
        else:
            ans = "Func"
        handler = True
        if attrs:
            attrs = attrs[2:-2]
            for attr in re.split(",", attrs):
                if attr == "block":
                    ans = "Blck"
                elif attr == "nohandler":
                    handler = False
                else:
                    raise AssertionError("Unknown attribute %s in: %s" % (attr, line))
        sim = Simcall(name, handler, Arg('result', ret), sargs, ans)
        if resdi is None:
            simcalls.append(sim)
        else:
            resdi.append(sim)
    return simcalls, simcalls_guarded


def header(name):
    fd = open(name, 'w')
    fd.write('/**********************************************************************/\n')
    fd.write('/* File generated by src/simix/simcalls.py from src/simix/simcalls.in */\n')
    fd.write('/*                                                                    */\n')
    fd.write('/*                    DO NOT EVER CHANGE THIS FILE                    */\n')
    fd.write('/*                                                                    */\n')
    fd.write('/* change simcalls specification in src/simix/simcalls.in             */\n')
    fd.write('/* Copyright (c) 2014-2022. The SimGrid Team. All rights reserved.    */\n')
    fd.write('/**********************************************************************/\n\n')
    fd.write('/*\n')
    fd.write(' * Note that the name comes from http://en.wikipedia.org/wiki/Popping\n')
    fd.write(' * Indeed, the control flow is doing a strange dance in there.\n')
    fd.write(' *\n')
    fd.write(' * That\'s not about http://en.wikipedia.org/wiki/Poop, despite the odor :)\n')
    fd.write(' */\n\n')
    return fd


def handle(fd, func, simcalls, guarded_simcalls):
    def nonempty(e):
        return e != ''
    fd.write('\n'.join(filter(nonempty, (func(simcall) for simcall in simcalls))))

    for guard, ll in guarded_simcalls.items():
        fd.write('\n#if %s\n' % (guard))
        fd.write('\n'.join(func(simcall) for simcall in ll))
        fd.write('\n#endif')

    fd.write('\n')

if __name__ == '__main__':
    simcalls, simcalls_dict = parse('simcalls.in')

    ok = True
    ok &= all(map(Simcall.check, simcalls))
    for k, v in simcalls_dict.items():
        ok &= all(map(Simcall.check, v))
    if not ok:
      print ("Some checks fail!")
      sys.exit(1)

    #
    # popping_accessors.hpp
    #
    fd = header('popping_accessors.hpp')
    fd.write('#include "src/simix/popping_private.hpp"')
    handle(fd, Simcall.accessors, simcalls, simcalls_dict)
    fd.write(
        "\n/* The prototype of all simcall handlers, automatically generated for you */\n\n")
    handle(fd, Simcall.handler_prototype, simcalls, simcalls_dict)
    fd.close()

    #
    # popping_enum.hpp
    #
    fd = header("popping_enum.hpp")
    fd.write('namespace simgrid {\n')
    fd.write('namespace simix {\n')
    fd.write('/**\n')
    fd.write(' * @brief All possible simcalls.\n')
    fd.write(' */\n')
    fd.write('enum class Simcall {\n')
    fd.write('  NONE,\n')

    handle(fd, Simcall.enum, simcalls, simcalls_dict)

    fd.write('};\n')
    fd.write('\n')
    fd.write('constexpr int NUM_SIMCALLS = ' + str(1 + len(simcalls)) + ';\n')
    fd.write('} // namespace simix\n')
    fd.write('} // namespace simgrid\n')
    fd.close()

    #
    # popping_generated.cpp
    #

    fd = header("popping_generated.cpp")

    fd.write('#include <simgrid/config.h>\n')
    fd.write('#include <simgrid/host.h>\n')
    fd.write('#include <xbt/base.h>\n')
    fd.write('#if SIMGRID_HAVE_MC\n')
    fd.write('#include "src/mc/mc_forward.hpp"\n')
    fd.write('#endif\n')
    fd.write('#include "src/kernel/activity/ConditionVariableImpl.hpp"\n')
    fd.write('#include "src/kernel/actor/SimcallObserver.hpp"\n')
    fd.write('#include "src/kernel/context/Context.hpp"\n')

    fd.write('\n')
    fd.write('XBT_LOG_EXTERNAL_DEFAULT_CATEGORY(simix);\n\n')

    fd.write('using simgrid::simix::Simcall;')
    fd.write('\n')
    fd.write('/** @brief Simcalls\' names (generated from src/simix/simcalls.in) */\n')
    fd.write('constexpr std::array<const char*, simgrid::simix::NUM_SIMCALLS> simcall_names{{\n')

    fd.write('    "Simcall::NONE",\n')
    handle(fd, Simcall.string, simcalls, simcalls_dict)

    fd.write('}};\n\n')

    fd.write('/** @private\n')
    fd.write(' * @brief (in kernel mode) unpack the simcall and activate the handler\n')
    fd.write(' *\n')
    fd.write(' * This function is generated from src/simix/simcalls.in\n')
    fd.write(' */\n')
    fd.write('void simgrid::kernel::actor::ActorImpl::simcall_handle(int times_considered)\n')
    fd.write('{\n')
    fd.write('  XBT_DEBUG("Handling simcall %p: %s", &simcall_, SIMIX_simcall_name(simcall_));\n')
    fd.write('  simcall_.mc_value_ = times_considered;\n')
    fd.write('  if (simcall_.observer_ != nullptr)\n')
    fd.write('    simcall_.observer_->prepare(times_considered);\n')

    fd.write('  if (context_->wannadie())\n')
    fd.write('    return;\n')
    fd.write('  switch (simcall_.call_) {\n')

    handle(fd, Simcall.case, simcalls, simcalls_dict)

    fd.write('    case Simcall::NONE:\n')
    fd.write('      throw std::invalid_argument(simgrid::xbt::string_printf("Asked to do the noop syscall on %s@%s",\n')
    fd.write('                                                              get_cname(),\n')
    fd.write('                                                              sg_host_get_name(get_host())));\n')
    fd.write('    default:\n')
    fd.write('      THROW_IMPOSSIBLE;\n')
    fd.write('  }\n')
    fd.write('}\n')

    fd.close()

    #
    # popping_bodies.cpp
    #
    fd = header('popping_bodies.cpp')
    fd.write('#include "src/kernel/EngineImpl.hpp"\n')
    fd.write('#include "src/kernel/actor/ActorImpl.hpp"\n')
    fd.write('#include "src/mc/mc_forward.hpp"\n')
    fd.write('#include "xbt/ex.h"\n')
    fd.write('#include <functional>\n')
    fd.write('#include <simgrid/simix.hpp>\n')
    fd.write('#include <xbt/log.h>\n')

    fd.write("/** @cond */ // Please Doxygen, don't look at this\n")
    fd.write('''
XBT_LOG_EXTERNAL_DEFAULT_CATEGORY(simix);

using simgrid::simix::Simcall;

template<class R, class... T>
inline static R simcall(Simcall call, T const&... t)
{
  auto self = simgrid::kernel::actor::ActorImpl::self();
  simgrid::simix::marshal(&self->simcall_, call, t...);
  if (not simgrid::kernel::EngineImpl::get_instance()->is_maestro(self)) {
    XBT_DEBUG("Yield process '%s' on simcall %s", self->get_cname(), SIMIX_simcall_name(self->simcall_));
    self->yield();
  } else {
    self->simcall_handle(0);
  }
  return simgrid::simix::unmarshal<R>(self->simcall_.result_);
}
''')
    handle(fd, Simcall.body, simcalls, simcalls_dict)
    fd.write("/** @endcond */\n")
    fd.close()



# Copyright (c) 2017-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from simgrid import Actor, Engine, Host, this_actor
import sys

# This example does not much: It just spans over-polite actor that yield a large amount
# of time before ending.
#
# This serves as an example for the simgrid.yield() function, with which an actor can request
# to be rescheduled after the other actor that are ready at the current timestamp.
#
# It can also be used to benchmark our context-switching mechanism.


def yielder (number_of_yields):
    for _ in range(number_of_yields):
        this_actor.yield_()
    this_actor.info("I yielded {:d} times. Goodbye now!".format(number_of_yields))

if __name__ == '__main__':
    e = Engine(sys.argv)

    e.load_platform(sys.argv[1])             # Load the platform description
  
    Actor.create("yielder", Host.by_name("Tremblay"), yielder, 10)
    Actor.create("yielder", Host.by_name("Ruby"), yielder, 15)

    e.run()  # - Run the simulation



# Copyright (c) 2017-2022. The SimGrid Team. All rights reserved.
#
# This program is free software you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from simgrid import Actor, Engine, Host, this_actor
import sys


def worker():
    """The worker actor, working for a while before leaving"""
    this_actor.info("Let's do some work (for 10 sec on Boivin).")
    this_actor.execute(980.95e6)

    this_actor.info("I'm done now. I leave even if it makes the daemon die.")


def my_daemon():
    """The daemon, displaying a message every 3 seconds until all other actors stop"""
    Actor.self().daemonize()

    while True:
        this_actor.info("Hello from the infinite loop")
        this_actor.sleep_for(3.0)

    this_actor.info(
        "I will never reach that point: daemons are killed when regular actors are done")


if __name__ == '__main__':
    e = Engine(sys.argv)
    if len(sys.argv) < 2:
        raise AssertionError(
            "Usage: actor-daemon.py platform_file [other parameters]")

    e.load_platform(sys.argv[1])
    Actor.create("worker", Host.by_name("Boivin"), worker)
    Actor.create("daemon", Host.by_name("Tremblay"), my_daemon)

    e.run()



## Copyright (c) 2007-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.
 
from simgrid import Actor, Engine, Host, Mailbox, this_actor, NetworkFailureException, TimeoutException
import sys

# This example shows how to work with the state profile of a host or a link,
# specifying when the resource must be turned on or off.
#
# To set such a profile, the first way is to use a file in the XML, while the second is to use the programmatic
# interface, as exemplified in the main() below. Once this profile is in place, the resource will automatically
# be turned on and off.
#
# The actors running on a host that is turned off are forcefully killed
# once their on_exit callbacks are executed. They cannot avoid this fate.
# Since we specified on_failure="RESTART" for each actors in the XML file,
# they will be automatically restarted when the host starts again.
#
# Communications using failed links will .. fail.

def master(* args):
  assert len(args) == 4, f"Actor master requires 4 parameters, but got {len(args)} ones."
  tasks_count   = int(args[0])
  comp_size     = int(args[1])
  comm_size     = int(args[2])
  workers_count = int(args[3])

  this_actor.info(f"Got {workers_count} workers and {tasks_count} tasks to process")

  for i in range(tasks_count): # For each task to be executed: 
    # - Select a worker in a round-robin way
    mailbox = Mailbox.by_name(f"worker-{i % workers_count}")
    try:
      this_actor.info(f"Send a message to {mailbox.name}")
      mailbox.put(comp_size, comm_size, 10.0)
      this_actor.info(f"Send to {mailbox.name} completed")
    except TimeoutException:
      this_actor.info(f"Mmh. Got timeouted while speaking to '{mailbox.name}'. Nevermind. Let's keep going!")
    except NetworkFailureException:
      this_actor.info(f"Mmh. The communication with '{mailbox.name}' failed. Nevermind. Let's keep going!")

  this_actor.info("All tasks have been dispatched. Let's tell everybody the computation is over.")
  for i in range (workers_count):
    # - Eventually tell all the workers to stop by sending a "finalize" task
    mailbox = Mailbox.by_name(f"worker-{i % workers_count}")
    try:
      mailbox.put(-1.0, 0, 1.0)
    except TimeoutException:
      this_actor.info(f"Mmh. Got timeouted while speaking to '{mailbox.name}'. Nevermind. Let's keep going!")
    except NetworkFailureException:
      this_actor.info(f"Mmh. The communication with '{mailbox.name}' failed. Nevermind. Let's keep going!")

  this_actor.info("Goodbye now!")

def worker(* args):
  assert len(args) == 1, "Expecting one parameter"
  my_id               = int(args[0])

  mailbox = Mailbox.by_name(f"worker-{my_id}")
  done = False
  while not done:
    try:
      this_actor.info(f"Waiting a message on {mailbox.name}")
      compute_cost = mailbox.get()
      if compute_cost > 0: # If compute_cost is valid, execute a computation of that cost 
        this_actor.info("Start execution...")
        this_actor.execute(compute_cost)
        this_actor.info("Execution complete.")
      else: # Stop when receiving an invalid compute_cost
        this_actor.info("I'm done. See you!")
        done = True
    except NetworkFailureException:
      this_actor.info("Mmh. Something went wrong. Nevermind. Let's keep going!")

def sleeper():
  this_actor.info("Start sleeping...")
  this_actor.sleep_for(1)
  this_actor.info("done sleeping.")

if __name__ == '__main__':
  assert len(sys.argv) > 2, f"Usage: python app-masterworkers.py platform_file deployment_file"

  e = Engine(sys.argv)

  # This is how to attach a profile to an host that is created from the XML file.
  # This should be done before calling load_platform(), as the on_creation() event is fired when loading the platform.
  # You can never set a new profile to a resource that already have one.
  def on_creation(host):
    if (host.name == "Bourrassa"):
      host.set_state_profile("67 0\n70 1\n", 0)
  Host.on_creation_cb(on_creation)

  e.load_platform(sys.argv[1])

  e.register_actor("master", master)
  e.register_actor("worker", worker)
  e.load_deployment(sys.argv[2])

  # Add a new host programatically, and attach a state profile to it
  lili = e.netzone_root.create_host("Lilibeth", 1e15)
  lili.set_state_profile("4 0\n5 1\n", 10)
  lili.seal()

  # Create an actor on that new host, to monitor its own state
  actor = Actor.create("sleeper", lili, sleeper)
  actor.set_auto_restart(True)

  e.run()

  this_actor.info(f"Simulation time {e.clock:.4f}")



# Copyright (c) 2017-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from simgrid import Actor, Engine, Host, this_actor
import sys


def victim_a_fun():
    this_actor.on_exit(lambda: this_actor.info("I have been killed!"))
    this_actor.info("Hello!")
    this_actor.info("Suspending myself")
    this_actor.suspend()                        # - Start by suspending itself
    # - Then is resumed and start to execute a task
    this_actor.info("OK, OK. Let's work")
    this_actor.execute(1e9)
    # - But will never reach the end of it
    this_actor.info("Bye!")


def victim_b_fun():
    this_actor.info("Terminate before being killed")


def killer():
    this_actor.info("Hello!")  # - First start a victim actor
    victim_a = Actor.create("victim A", Host.by_name("Fafard"), victim_a_fun)
    victim_b = Actor.create("victim B", Host.by_name("Jupiter"), victim_b_fun)
    this_actor.sleep_for(10)  # - Wait for 10 seconds

    # - Resume it from its suspended state
    this_actor.info("Resume the victim A")
    victim_a.resume()
    this_actor.sleep_for(2)

    this_actor.info("Kill the victim A")   # - and then kill it
    Actor.by_pid(victim_a.pid).kill()       # You can retrieve an actor from its PID (and then kill it)

    this_actor.sleep_for(1)

    # that's a no-op, there is no zombies in SimGrid
    this_actor.info("Kill victim B, even if it's already dead")
    victim_b.kill()

    this_actor.sleep_for(1)

    this_actor.info("Start a new actor, and kill it right away")
    victim_c = Actor.create("victim C", Host.by_name("Jupiter"), victim_a_fun)
    victim_c.kill()

    this_actor.sleep_for(1)

    this_actor.info("Killing everybody but myself")
    Actor.kill_all()

    this_actor.info("OK, goodbye now. I commit a suicide.")
    this_actor.exit()

    this_actor.info(
        "This line never gets displayed: I'm already dead since the previous line.")


if __name__ == '__main__':
    e = Engine(sys.argv)
    if len(sys.argv) < 2:
        raise AssertionError(
            "Usage: actor-kill.py platform_file [other parameters]")

    e.load_platform(sys.argv[1])     # Load the platform description
    # Create and deploy killer actor, that will create the victim actors
    Actor.create("killer", Host.by_name("Tremblay"), killer)

    e.run()



#!/usr/bin/env python

# Copyright (c) 2012-2022. The SimGrid Team.
# All rights reserved.

# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

# This script generates a specific deployment file for the Bittorrent example.
# It assumes that the platform will be a cluster.
# Usage: python generate.py nb_nodes nb_bits end_date percentage
# Example: python generate.py 10000 5000

import sys
import random

if len(sys.argv) != 4:
    print(
        "Usage: python generate.py nb_nodes end_date seed_percentage > deployment_file.xml")
    sys.exit(1)

nb_nodes = int(sys.argv[1])
end_date = int(sys.argv[2])
seed_percentage = int(sys.argv[3])

nb_bits = 24
max_id = 2 ** nb_bits - 1
all_ids = [42]

sys.stdout.write("<?xml version='1.0'?>\n"
                 "<!DOCTYPE platform SYSTEM \"https://simgrid.org/simgrid.dtd\">\n"
                 "<platform version=\"4\">\n"
                 "  <actor host=\"node-0.simgrid.org\" function=\"tracker\">\n"
                 "    <argument value=\"%d\"/>\n  </actor>\n" % end_date)

for i in range(1, nb_nodes):

    ok = False
    while not ok:
        my_id = random.randint(0, max_id)
        ok = my_id not in all_ids
    start_date = i * 10
    line = "  <actor host=\"node-%d.simgrid.org\" function=\"peer\">\n" % i
    line += "    <argument value=\"%d\"/>\n    <argument value=\"%d\"/>\n" % (
        my_id, end_date)
    if random.randint(0, 100) < seed_percentage:
        line += "    <argument value=\"1\"/>\n"
    line += "  </actor>\n"
    sys.stdout.write(line)
    all_ids.append(my_id)
sys.stdout.write("</platform>")



#
# Copyright 2012-2015 Bronto Software, Inc. and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import javalang

import javasphinx.formatter as formatter
import javasphinx.util as util
import javasphinx.htmlrst as htmlrst

class JavadocRestCompiler(object):
    """ Javadoc to ReST compiler. Builds ReST documentation from a Java syntax
    tree. """

    def __init__(self, filter=None, member_headers=True, parser='lxml'):
        if filter:
            self.filter = filter
        else:
            self.filter = self.__default_filter

        self.converter = htmlrst.Converter(parser)

        self.member_headers = member_headers

    def __default_filter(self, node):
        """Excludes private members and those tagged "@hide" / "@exclude" in their
        docblocks.

        """

        if not isinstance(node, javalang.tree.Declaration):
            return False

        if 'private' in node.modifiers:
            return False

        if isinstance(node, javalang.tree.Documented) and node.documentation:
            doc = javalang.javadoc.parse(node.documentation)
            if 'hide' in doc.tags or 'exclude' in doc.tags:
                return False

        return True

    def __html_to_rst(self, s):
        return self.converter.convert(s)

    def __output_doc(self, documented):
        if not isinstance(documented, javalang.tree.Documented):
            raise ValueError('node not documented')

        output = util.Document()

        if not documented.documentation:
            return output

        doc = javalang.javadoc.parse(documented.documentation)

        if doc.description:
            output.add(self.__html_to_rst(doc.description))
            output.clear()

        if doc.authors:
            output.add_line(':author: %s' % (self.__html_to_rst(', '.join(doc.authors)),))

        for name, value in doc.params:
            output.add_line(':param %s: %s' % (name, self.__html_to_rst(value)))

        for exception in doc.throws:
            description = doc.throws[exception]
            output.add_line(':throws %s: %s' % (exception, self.__html_to_rst(description)))

        if doc.return_doc:
            output.add_line(':return: %s' % (self.__html_to_rst(doc.return_doc),))

        if doc.tags.get('see'):
            output.clear()

            see_also = ', '.join(self.__output_see(see) for see in doc.tags['see'])
            output.add_line('**See also:** %s' % (see_also,))

        return output

    def __output_see(self, see):
        """ Convert the argument to a @see tag to rest """

        if see.startswith('<a href'):
            # HTML link -- <a href="...">...</a>
            return self.__html_to_rst(see)
        elif '"' in see:
            # Plain text
            return see
        else:
            # Type reference (default)
            return ':java:ref:`%s`' % (see.replace('#', '.').replace(' ', ''),)

    def compile_type(self, declaration):
        signature = util.StringBuilder()
        formatter.output_declaration(declaration, signature)

        doc = self.__output_doc(declaration)

        directive = util.Directive('java:type', signature.build())
        directive.add_content(doc)

        return directive

    def compile_enum_constant(self, enum, constant):
        signature = util.StringBuilder()

        for annotation in constant.annotations:
            formatter.output_annotation(annotation, signature)

        # All enum constants are public, static, and final
        signature.append('public static final ')
        signature.append(enum)
        signature.append(' ')
        signature.append(constant.name)

        doc = self.__output_doc(constant)

        directive = util.Directive('java:field', signature.build())
        directive.add_content(doc)

        return directive

    def compile_field(self, field):
        signature = util.StringBuilder()

        for annotation in field.annotations:
            formatter.output_annotation(annotation, signature)

        formatter.output_modifiers(field.modifiers, signature)
        signature.append(' ')

        formatter.output_type(field.type, signature)
        signature.append(' ')
        signature.append(field.declarators[0].name)

        doc = self.__output_doc(field)

        directive = util.Directive('java:field', signature.build())
        directive.add_content(doc)

        return directive

    def compile_constructor(self, constructor):
        signature = util.StringBuilder()

        for annotation in constructor.annotations:
            formatter.output_annotation(annotation, signature)

        formatter.output_modifiers(constructor.modifiers, signature)
        signature.append(' ')

        if constructor.type_parameters:
            formatter.output_type_params(constructor.type_parameters, signature)
            signature.append(' ')

        signature.append(constructor.name)

        signature.append('(')
        formatter.output_list(formatter.output_formal_param, constructor.parameters, signature, ', ')
        signature.append(')')

        if constructor.throws:
            signature.append(' throws ')
            formatter.output_list(formatter.output_exception, constructor.throws, signature, ', ')

        doc = self.__output_doc(constructor)

        directive = util.Directive('java:constructor', signature.build())
        directive.add_content(doc)

        return directive

    def compile_method(self, method):
        signature = util.StringBuilder()

        for annotation in method.annotations:
            formatter.output_annotation(annotation, signature)

        formatter.output_modifiers(method.modifiers, signature)
        signature.append(' ')

        if method.type_parameters:
            formatter.output_type_params(method.type_parameters, signature)
            signature.append(' ')

        formatter.output_type(method.return_type, signature)
        signature.append(' ')

        signature.append(method.name)

        signature.append('(')
        formatter.output_list(formatter.output_formal_param, method.parameters, signature, ', ')
        signature.append(')')

        if method.throws:
            signature.append(' throws ')
            formatter.output_list(formatter.output_exception, method.throws, signature, ', ')

        doc = self.__output_doc(method)

        directive = util.Directive('java:method', signature.build())
        directive.add_content(doc)

        return directive

    def compile_type_document(self, imports_block, package, name, declaration):
        """ Compile a complete document, documenting a type and its members """

        outer_type = name.rpartition('.')[0]

        document = util.Document()
        document.add(imports_block)
        document.add_heading(name, '=')

        method_summary = util.StringBuilder()
        document.add_object(method_summary)

        package_dir = util.Directive('java:package', package)
        package_dir.add_option('noindex')
        document.add_object(package_dir)

        # Add type-level documentation
        type_dir = self.compile_type(declaration)
        if outer_type:
            type_dir.add_option('outertype', outer_type)
        document.add_object(type_dir)

        if isinstance(declaration, javalang.tree.EnumDeclaration):
            enum_constants = list(declaration.body.constants)
            enum_constants.sort(key=lambda c: c.name)

            document.add_heading('Enum Constants')
            for enum_constant in enum_constants:
                if self.member_headers:
                    document.add_heading(enum_constant.name, '^')
                c = self.compile_enum_constant(name, enum_constant)
                c.add_option('outertype', name)
                document.add_object(c)

        fields = list(filter(self.filter, declaration.fields))
        if fields:
            document.add_heading('Fields', '-')
            fields.sort(key=lambda f: f.declarators[0].name)
            for field in fields:
                if self.member_headers:
                    document.add_heading(field.declarators[0].name, '^')
                f = self.compile_field(field)
                f.add_option('outertype', name)
                document.add_object(f)

        constructors = list(filter(self.filter, declaration.constructors))
        if constructors:
            document.add_heading('Constructors', '-')
            constructors.sort(key=lambda c: c.name)
            for constructor in constructors:
                if self.member_headers:
                    document.add_heading(constructor.name, '^')
                c = self.compile_constructor(constructor)
                c.add_option('outertype', name)
                document.add_object(c)

        methods = list(filter(self.filter, declaration.methods))
        if methods:
            document.add_heading('Methods', '-')
            methods.sort(key=lambda m: m.name)
            for method in methods:
                if self.member_headers:
                    document.add_heading(method.name, '^')
                m = self.compile_method(method)
                m.add_option('outertype', name)
                document.add_object(m)

        return document

    def compile(self, ast):
        """ Compile autodocs for the given Java syntax tree. Documents will be
        returned documenting each separate type. """

        documents = {}

        imports = util.StringBuilder()
        for imp in ast.imports:
            if imp.static or imp.wildcard:
                continue

            package_parts = []
            cls_parts = []

            for part in imp.path.split('.'):
                if cls_parts or part[0].isupper():
                    cls_parts.append(part)
                else:
                    package_parts.append(part)


            # If the import's final part wasn't capitalized,
            # append it to the class parts anyway so sphinx doesn't complain.
            if cls_parts == []:
                cls_parts.append(package_parts.pop())

            package = '.'.join(package_parts)
            cls = '.'.join(cls_parts)

            imports.append(util.Directive('java:import', package + ' ' + cls).build())
        import_block = imports.build()

        if not ast.package:
            raise ValueError('File must have package declaration')

        package = ast.package.name
        type_declarations = []
        for path, node in ast.filter(javalang.tree.TypeDeclaration):
            if not self.filter(node):
                continue

            classes = [n.name for n in path if isinstance(n, javalang.tree.TypeDeclaration)]
            classes.append(node.name)

            name = '.'.join(classes)
            type_declarations.append((package, name, node))

        for package, name, declaration in type_declarations:
            full_name = package + '.' + name
            document = self.compile_type_document(import_block, package, name, declaration)
            documents[full_name] = (package, name, document.build())
        return documents

    def compile_docblock(self, documented):
        ''' Compiles a single, standalone docblock. '''
        return self.__output_doc(documented).build()



# Copyright (c) 2017-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

# This example demonstrate the actor migrations.
#
# The worker actor first move by itself, and then start an execution.
# During that execution, the monitor migrates the worker, that wakes up on another host.
# The execution was of the right amount of flops to take exactly 5 seconds on the first host
# and 5 other seconds on the second one, so it stops after 10 seconds.
#
# Then another migration is done by the monitor while the worker is suspended.
#
# Note that worker() takes an uncommon set of parameters,
# and that this is perfectly accepted by create().

from simgrid import Actor, Engine, Host, this_actor
import sys


def worker(first_host, second_host):
    flop_amount = first_host.speed * 5 + second_host.speed * 5

    this_actor.info("Let's move to {:s} to execute {:.2f} Mflops (5sec on {:s} and 5sec on {:s})".format(
        first_host.name, flop_amount / 1e6, first_host.name, second_host.name))

    this_actor.set_host(first_host)
    this_actor.execute(flop_amount)

    this_actor.info("I wake up on {:s}. Let's suspend a bit".format(
        this_actor.get_host().name))

    this_actor.suspend()

    this_actor.info("I wake up on {:s}".format(this_actor.get_host().name))
    this_actor.info("Done")


def monitor():
    boivin = Host.by_name("Boivin")
    jacquelin = Host.by_name("Jacquelin")
    fafard = Host.by_name("Fafard")

    actor = Actor.create("worker", fafard, worker, boivin, jacquelin)

    this_actor.sleep_for(5)

    this_actor.info(
        "After 5 seconds, move the actor to {:s}".format(jacquelin.name))
    actor.host = jacquelin

    this_actor.sleep_until(15)
    this_actor.info(
        "At t=15, move the actor to {:s} and resume it.".format(fafard.name))
    actor.host = fafard
    actor.resume()


if __name__ == '__main__':
    e = Engine(sys.argv)
    if len(sys.argv) < 2:
        raise AssertionError(
            "Usage: actor-migration.py platform_file [other parameters]")
    e.load_platform(sys.argv[1])

    Actor.create("monitor", Host.by_name("Boivin"), monitor)
    e.run()



# Copyright (c) 2007-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

# This Python file acts as the foil to the corresponding XML file, where the
# action takes place: Actors are started and stopped at predefined time

from simgrid import Engine, this_actor
import sys


class Sleeper:
    """This actor just sleeps until termination"""

    def __init__(self, *args):
        this_actor.on_exit(lambda: this_actor.info("Exiting now (done sleeping or got killed)."))

    def __call__(self):
        this_actor.info("Hello! I go to sleep.")
        this_actor.sleep_for(10)
        this_actor.info("Done sleeping.")


if __name__ == '__main__':
    e = Engine(sys.argv)
    if len(sys.argv) < 2:
        raise AssertionError(
            "Usage: actor-lifetime.py platform_file actor-lifetime_d.xml [other parameters]")

    e.load_platform(sys.argv[1])     # Load the platform description
    e.register_actor("sleeper", Sleeper)
    # Deploy the sleeper actors with explicit start/kill times
    e.load_deployment(sys.argv[2])

    e.run()



# Copyright (c) 2007-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from simgrid import Actor, Engine, Host, this_actor
import sys


class Dvfs:
    def __call__(self):
        workload = 100E6
        host = this_actor.get_host()

        nb = host.pstate_count
        this_actor.info("Count of Processor states={:d}".format(nb))

        this_actor.info("Current power peak={:f}".format(host.speed))

        # Run a task
        this_actor.execute(workload)

        task_time = Engine.clock
        this_actor.info("Task1 duration: {:.2f}".format(task_time))

        # Change power peak
        new_pstate = 2

        this_actor.info("Changing power peak value to {:f} (at index {:d})".format( host.pstate_speed(new_pstate), new_pstate))

        host.pstate = new_pstate

        this_actor.info("Changed power peak={:f}".format(host.speed))

        # Run a second task
        this_actor.execute(workload)

        task_time = Engine.clock - task_time
        this_actor.info("Task2 duration: {:.2f}".format(task_time))

        # Verify that the default pstate is set to 0
        host2 = Host.by_name("MyHost2")
        this_actor.info("Count of Processor states={:d}".format(host2.pstate_count))

        this_actor.info("Final power peak={:f}".format(host2.speed))

if __name__ == '__main__':
    e = Engine(sys.argv)
    if len(sys.argv) < 2:
        raise AssertionError("Usage: exec-dvfs.py platform_file [other parameters] (got {:d} params)".format(len(sys.argv)))

    e.load_platform(sys.argv[1])
    Actor.create("dvfs_test", Host.by_name("MyHost1"), Dvfs())
    Actor.create("dvfs_test", Host.by_name("MyHost2"), Dvfs())

    e.run()



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.

#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from simgrid import Actor, Engine, Host, Mailbox, this_actor
import sys

# This example shows how to use simgrid::s4u::this_actor::wait() to wait for a given communication.
#
# As for the other asynchronous examples, the sender initiate all the messages it wants to send and
# pack the resulting simgrid::s4u::CommPtr objects in a vector. All messages thus occurs concurrently.
#
# The sender then loops until there is no ongoing communication.


def sender(messages_count, msg_size, receivers_count):
     # List in which we store all ongoing communications
     pending_comms = []

     # Vector of the used mailboxes
     mboxes = [Mailbox.by_name("receiver-{:d}".format(i)) for i in range(0, receivers_count)]

     # Start dispatching all messages to receivers, in a round robin fashion
     for i in range(0, messages_count):
         content = "Message {:d}".format(i)
         mbox = mboxes[i % receivers_count]

         this_actor.info("Send '{:s}' to '{:s}'".format(content, str(mbox)))

         # Create a communication representing the ongoing communication, and store it in pending_comms
         comm = mbox.put_async(content, msg_size)
         pending_comms.append(comm)

     # Start sending messages to let the workers know that they should stop
     for i in range(0, receivers_count):
         mbox = mboxes[i]
         this_actor.info("Send 'finalize' to '{:s}'".format(str(mbox)))
         comm = mbox.put_async("finalize", 0)
         pending_comms.append(comm)

     this_actor.info("Done dispatching all messages")

     # Now that all message exchanges were initiated, wait for their completion, in order of creation.
     for comm in pending_comms:
         comm.wait()
     this_actor.info("Goodbye now!")

def receiver(id):
     mbox = Mailbox.by_name("receiver-{:d}".format(id))
     this_actor.info("Wait for my first message")
     while True:
         received = mbox.get()
         this_actor.info("I got a '{:s}'.".format(received))
         if received == "finalize":
             break  # If it's a finalize message, we're done.


if __name__ == '__main__':
    e = Engine(sys.argv)

    e.load_platform(sys.argv[1])             # Load the platform description

    Actor.create("sender", Host.by_name("Tremblay"), sender, 3, 50000000, 1)
    Actor.create("receiver", Host.by_name("Ruby"), receiver, 0)
  
    e.run()



# Copyright (c) 2019-2022. The SimGrid Team. All rights reserved.

# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

# python3 setup.py sdist # Build a source distrib (building binary distribs is complex on linux)

# twine upload --repository-url https://test.pypi.org/legacy/ dist/simgrid-*.tar.gz # Upload to test
# pip3 install --user --index-url https://test.pypi.org/simple  simgrid

# Once it works, upload to the real infra.  /!\ you cannot modify a file once uploaded
# twine upload dist/simgrid-*.tar.gz

import os
import re
import sys
import platform
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build python bindings of SimGrid")

        if not os.path.exists("MANIFEST.in"):
            raise RuntimeError(
                "Please generate a MANIFEST.in file (configure simgrid, and copy it here if you build out of tree)")

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        from pybind11 import get_cmake_dir
        extdir = os.path.abspath(os.path.dirname(
            self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable,
                      '-Denable_smpi=OFF',
                      '-Denable_java=OFF',
                      '-Denable_python=ON',
                      '-Dminimal-bindings=ON',
                      '-Dpybind11_DIR=' + get_cmake_dir()
                      ]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += [
                '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j4']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] +
                              cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] +
                              build_args, cwd=self.build_temp)


setup(
    name='simgrid',
    version='3.30',
    author='Da SimGrid Team',
    author_email='simgrid-devel@lists.gforge.inria.fr',
    description='Toolkit for scalable simulation of distributed applications',
    long_description=("SimGrid is a scientific instrument to study the behavior of "
                      "large-scale distributed systems such as Grids, Clouds, HPC or P2P "
                      "systems. It can be used to evaluate heuristics, prototype applications "
                      "or even assess legacy MPI applications.\n\n"
                      "This package contains a native library. Please install cmake, boost, pybind11 and a "
                      "C++ compiler before using pip3. On Debian/Ubuntu, this is as easy as\n"
                      "sudo apt install cmake libboost-dev pybind11-dev g++ gcc"),
    ext_modules=[CMakeExtension('simgrid')],
    cmdclass=dict(build_ext=CMakeBuild),
    install_requires=['pybind11>=2.4'],
    setup_requires=['pybind11>=2.4'],
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        "Programming Language :: C++",
        "Programming Language :: C",
        "Programming Language :: Fortran",
        "Programming Language :: Java",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Systems Administration",
    ],
    url="https://simgrid.org",
    project_urls={
        'Tracker': 'https://framagit.org/simgrid/simgrid/issues/',
        'Source':  'https://framagit.org/simgrid/simgrid/',
        'Documentation': 'https://simgrid.org/doc/latest/',
    },
)



# Copyright (c) 2018-2022. The SimGrid Team. All rights reserved.
#
# This program is free software you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from simgrid import Actor, Engine, Host, this_actor
import sys


class Wizard:
    def __call__(self):

        fafard = Host.by_name("Fafard")
        ginette = Host.by_name("Ginette")
        boivin = Host.by_name("Boivin")

        this_actor.info("I'm a wizard! I can run a task on the Ginette host from the Fafard one! Look!")
        activity = this_actor.exec_init(48.492e6)
        activity.host = ginette
        activity.start()
        this_actor.info("It started. Running 48.492Mf takes exactly one second on Ginette (but not on Fafard).")

        this_actor.sleep_for(0.1)
        this_actor.info("Loads in flops/s: Boivin={:.0f}; Fafard={:.0f}; Ginette={:.0f}".format(boivin.load, fafard.load,
                                                                                                ginette.load))
        activity.wait()
        this_actor.info("Done!")

        this_actor.info("And now, harder. Start a remote task on Ginette and move it to Boivin after 0.5 sec")
        activity = this_actor.exec_init(73293500)
        activity.host = ginette
        activity.start()

        this_actor.sleep_for(0.5)
        this_actor.info(
            "Loads before the move: Boivin={:.0f}; Fafard={:.0f}; Ginette={:.0f}".format(
                boivin.load,
                fafard.load,
                ginette.load))

        activity.host = boivin

        this_actor.sleep_for(0.1)
        this_actor.info(
            "Loads after the move: Boivin={:.0f}; Fafard={:.0f}; Ginette={:.0f}".format(
                boivin.load,
                fafard.load,
                ginette.load))

        activity.wait()
        this_actor.info("Done!")


if __name__ == '__main__':
    e = Engine(sys.argv)

    e.load_platform(sys.argv[1])

    Actor.create("test", Host.by_name("Fafard"), Wizard())

    e.run()



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from simgrid import Actor, Comm, Engine, Host, Mailbox, this_actor
import sys

# This example shows how to block on the completion of a set of communications.
#
# As for the other asynchronous examples, the sender initiate all the messages it wants to send and
# pack the resulting simgrid.Comm objects in a list. All messages thus occur concurrently.
#
# The sender then loops until there is no ongoing communication. Using wait_any() ensures that the sender
# will notice events as soon as they occur even if it does not follow the order of the container.
#
# Here, finalize messages will terminate earlier because their size is 0, so they travel faster than the
# other messages of this application.  As expected, the trace shows that the finalize of worker 1 is
# processed before 'Message 5' that is sent to worker 0.


def sender(messages_count, msg_size, receivers_count):
     # List in which we store all ongoing communications
     pending_comms = []

     # Vector of the used mailboxes
     mboxes = [Mailbox.by_name("receiver-{:d}".format(i))
               for i in range(0, receivers_count)]

     # Start dispatching all messages to receivers, in a round robin fashion
     for i in range(0, messages_count):
         content = "Message {:d}".format(i)
         mbox = mboxes[i % receivers_count]

         this_actor.info("Send '{:s}' to '{:s}'".format(content, str(mbox)))

         # Create a communication representing the ongoing communication, and store it in pending_comms
         comm = mbox.put_async(content, msg_size)
         pending_comms.append(comm)

     # Start sending messages to let the workers know that they should stop
     for i in range(0, receivers_count):
         mbox = mboxes[i]
         this_actor.info("Send 'finalize' to '{:s}'".format(str(mbox)))
         comm = mbox.put_async("finalize", 0)
         pending_comms.append(comm)

     this_actor.info("Done dispatching all messages")

     # Now that all message exchanges were initiated, wait for their completion, in order of completion.
     #
     # This loop waits for first terminating message with wait_any() and remove it with del, until all comms are
     # terminated.
     # Even in this simple example, the pending comms do not terminate in the exact same order of creation.
     while pending_comms:
         changed_pos = Comm.wait_any(pending_comms)
         del pending_comms[changed_pos]
         if (changed_pos != 0):
             this_actor.info(
                  "Remove the {:d}th pending comm: it terminated earlier than another comm that was initiated first.".format(changed_pos))

     this_actor.info("Goodbye now!")


def receiver(id):
    mbox = Mailbox.by_name("receiver-{:d}".format(id))
    this_actor.info("Wait for my first message")
    while True:
        received = mbox.get()
        this_actor.info("I got a '{:s}'.".format(received))
        if received == "finalize":
            break  # If it's a finalize message, we're done.

if __name__ == '__main__':
    e = Engine(sys.argv)

    # Load the platform description
    e.load_platform(sys.argv[1])

    Actor.create("sender", Host.by_name("Tremblay"), sender, 6, 1000000, 2)
    Actor.create("receiver", Host.by_name("Fafard"), receiver, 0)
    Actor.create("receiver", Host.by_name("Jupiter"), receiver, 1)

    e.run()
